@@ -1,6 +1,6 @@
 /*
  * balance - a balancing tcp proxy
- * $Revision: 3.52 $
+ * $Revision: 3.56 $
  *
  * Copyright (c) 2000-2009,2010 by Thomas Obermair (obermair@acm.org)
  * and Inlab Software GmbH (info@inlab.de), Gruenwald, Germany.
@@ -13,6 +13,11 @@
  *
  * This program is dedicated to Richard Stevens...
  *
+ *  3.56
+ *    added out-of-band data handling
+ *    thanks to Julian Griffiths
+ *  3.54
+ *    fixed hash_fold bug regarding incoming IPv4 and IPv6 source addresses
  *  3.52
  *    thanks to David J. Jilk from Standing Cloud, Inc. for the following:
  *    added "nobuffer" functionality to interactive shell IO
@@ -106,8 +111,8 @@
 
 #include <balance.h>
 
-const char *balance_rcsid = "$Id: balance.c,v 3.52 2010/02/14 18:35:09 t Exp $";
-static char *revision = "$Revision: 3.52 $";
+const char *balance_rcsid = "$Id: balance.c,v 3.56 2013/11/06 10:55:10 t Exp $";
+static char *revision = "$Revision: 3.56 $";
 
 static int release;
 static int subrelease;
@@ -118,6 +123,25 @@ static int rendezvousfd;
 static int shmfilefd;
 #endif
 
+static int cur_s;
+static int cur_r;
+
+static void  urg_handler(int signo) {
+	int n;
+	char buf[256];
+
+	n = recv(cur_r,buf,sizeof buf,MSG_OOB);
+	if ( n < 0 ) {
+		//printf("ERROR: recv(2)\n");
+	}else{
+
+		buf[n] = 0;
+		//printf("URG '%s' (%d)\n", buf,n);
+		send(cur_s, buf,strlen(buf),MSG_OOB);
+	}
+	signal(SIGURG, urg_handler);
+}
+
 static int err_dump(char *text) {
   fprintf(stderr, "balance: %s\n", text);
   fflush(stderr);
@@ -152,9 +176,16 @@ int create_serversocket(char* node, char* service) {
 
   bzero(&hints, sizeof(hints));
   hints.ai_flags = AI_PASSIVE;
+
   if(bindipv6) {
+    if(debugflag) {
+      fprintf(stderr, "using AF_INET6\n");
+    }
     hints.ai_family = AF_INET6;
   } else {
+    if(debugflag) {
+      fprintf(stderr, "using AF_UNSPEC\n");
+    }
     hints.ai_family = AF_UNSPEC;
   }
   hints.ai_socktype = SOCK_STREAM;
@@ -189,7 +220,9 @@ int create_serversocket(char* node, char* service) {
 #endif
 
   sockopton = 1;
+
   status = setsockopt(srv_socket, SOL_SOCKET, SO_REUSEADDR, (char*) &sockopton, sizeof(sockopton));
+
   if(status < 0) {
     perror("setsockopt(SO_REUSEADDR=1)");
     exit(EX_OSERR);
@@ -580,7 +613,23 @@ int forward(int fromfd, int tofd, int groupindex, int channelindex)
   ssize_t rc;
   unsigned char buffer[MAXTXSIZE];
 
-  rc = read(fromfd, buffer, MAXTXSIZE);
+	cur_s = tofd;
+	cur_r = fromfd;
+/*
+  struct sigaction urg_action;
+
+    urg_action.sa_handler = urg_handler;
+    urg_action.sa_flags = SA_RESTART;
+    sigemptyset(&urg_action.sa_mask);
+    sigaction(SIGURG, &urg_action, NULL);
+*/
+	signal(SIGURG, &urg_handler);
+
+
+
+  //rc = read(fromfd, buffer, MAXTXSIZE);
+  fcntl(fromfd, F_SETOWN,getpid());
+  rc = recv(fromfd, buffer, MAXTXSIZE, 0);
 
   if (packetdump) {
     printf("-> %d\n", (int) rc);
@@ -604,7 +653,6 @@ int backward(int fromfd, int tofd, int groupindex, int channelindex)
 {
   ssize_t rc;
   unsigned char buffer[MAXTXSIZE];
-
   rc = read(fromfd, buffer, MAXTXSIZE);
 
   if (packetdump) {
@@ -700,6 +748,7 @@ void stream2(int clientfd, int serverfd, int groupindex, int channelindex)
   exit(EX_OK);
 }
 
+
 void alrm_handler(int signo) {
 }
 
@@ -767,6 +816,7 @@ void *stream(int arg, int groupindex, int index, char *client_address,
     serv_addr.sin_port = htons(chn_port(common, groupindex, index));
     b_unlock();
 
+
     alrm_action.sa_handler = alrm_handler;
     alrm_action.sa_flags = 0;	// don't restart !
     sigemptyset(&alrm_action.sa_mask);
@@ -857,10 +907,12 @@ void *stream(int arg, int groupindex, int index, char *client_address,
 	      break;
 	    } else if (grp_type(common, groupindex) == GROUP_HASH) {
 	      unsigned int uindex;
-	      uindex = hash_fold(client_address, client_address_size);
+	      uindex = hash_fold((unsigned char*) &(((struct sockaddr_in6 *) &client_address)->sin6_addr), client_address_size);
+
+	      if (debugflag) {
+		fprintf(stderr, "HASH-method: fold returns %u\n", uindex);
+              }
 
-	      if (debugflag)
-		fprintf(stderr, "fold returns %d\n", uindex);
 	      index = uindex % grp_nchannels(common, groupindex);
 	      if (debugflag)
 		fprintf(stderr, "modulo %d gives %d\n",
@@ -1505,7 +1557,7 @@ int main(int argc, char *argv[])
   int fd;
   char *argument = NULL;
   struct stat buffer;
-  struct sockaddr_in cli_addr;
+  struct sockaddr_storage cli_addr;
   struct sigaction usr1_action, chld_action;
 #ifdef BalanceBSD
 #else
@@ -1600,6 +1652,8 @@ int main(int argc, char *argv[])
       usage();
     }
   }
+
+
   usr1_action.sa_handler = usr1_handler;
   usr1_action.sa_flags = SA_RESTART;
   sigemptyset(&usr1_action.sa_mask);
@@ -1691,14 +1745,12 @@ int main(int argc, char *argv[])
 
   openlog("Balance", LOG_ODELAY | LOG_PID | LOG_CONS, LOG_DAEMON);
 
-  // Open a TCP socket (an Internet stream socket).
+  /*  Open a TCP socket (an Internet stream socket). */
 
   sockfd = create_serversocket(bindhost, argv[0]);
 
-  (void) setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &sockbufsize,
-    sizeof(sockbufsize));
-  (void) setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &sockbufsize,
-    sizeof(sockbufsize));
+  (void) setsockopt(sockfd, SOL_SOCKET, SO_SNDBUF, &sockbufsize, sizeof(sockbufsize));
+  (void) setsockopt(sockfd, SOL_SOCKET, SO_RCVBUF, &sockbufsize, sizeof(sockbufsize));
 
   // init of common (*after* bind()) 
 
@@ -1724,14 +1776,16 @@ int main(int argc, char *argv[])
     }
 
     if (debugflag) {
-      fprintf(stderr, "connect from %s\n", inet_ntoa(cli_addr.sin_addr));
+      char buf[1024];
+      inet_ntop(AF_INET6,&(((struct sockaddr_in6*) &cli_addr)->sin6_addr),buf,1024);
+      fprintf(stderr, "connect from %s clilen=%d\n", buf, clilen);
     }
 
     /* 
      * the balancing itself:
      * - groupindex = 0
      * - decision wich channel to use for the first try 
-     * - client address available in cli_addr.sin_addr
+     * - client address available in cli_addr
      *
      */
 
@@ -1756,9 +1810,12 @@ int main(int argc, char *argv[])
 	    }
 	  }
 	} else if (grp_type(common, groupindex) == GROUP_HASH) {
-	  uindex = hash_fold((char *) &cli_addr.sin_addr, sizeof(cli_addr.sin_addr));
-	  if (debugflag)
-	    fprintf(stderr, "fold returns %d\n", uindex);
+	  uindex = hash_fold((unsigned char*) &(((struct sockaddr_in6 *) &cli_addr)->sin6_addr), clilen);
+   
+	  if(debugflag) {
+	    fprintf(stderr, "HASH-method: fold returns %u\n", uindex);
+          }
+
 	  index = uindex % grp_nchannels(common, groupindex);
 	  if (debugflag)
 	    fprintf(stderr, "modulo %d gives %d\n",
@@ -1855,8 +1912,7 @@ int main(int argc, char *argv[])
 	close(sockfd);			// close original socket 
 	// process the request: 
 
-	stream(newsockfd, groupindex, index, (char *) &cli_addr.sin_addr,
-	       sizeof(cli_addr.sin_addr));
+	stream(newsockfd, groupindex, index, (char *) &cli_addr, clilen);
 	exit(EX_OK);
       }
     }
